﻿using System;
using System.Collections.Generic;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapters 6 and 7

    public class Matrix
    {
        public readonly int NRows;              //the number of rows in the matrix
        public readonly int NColumns;           //the number of columns in the matrix
        private readonly double[][] _elements;    //the elements of the matrix. Even though all rows have the same lenght, [][] is faster than [,] for many computations.

        #region Constructors
        /// <summary>
        /// Creates a new matrix with elements corresponding to the elements of the array.
        /// </summary>
        /// <param name="array">The elements of this array will be the initial elements of the matrix.</param>
        public Matrix(double[,] array)
        {
            NRows = array.GetLength(0);
            NColumns = array.GetLength(1);
            if(NRows == 0 || NColumns == 0)
                throw new ArgumentException("The elements of the array must have at least one row and one column");
            _elements = new double[NRows][];
            for (int r = 0; r < NRows; r++)
            {
                _elements[r] = new double[NColumns];
                for (int j = 0; j < NColumns; j++)
                    _elements[r][j] = array[r, j];
            }
        }

        /// <summary>
        /// Creates a vector, a matrix with one column.
        /// </summary>
        /// <param name="array">The elements of this array will be the initial elements of the matrix.</param>
        public Matrix(double[] array)
        {
            NRows = array.Length;
            if (NRows == 0)
                throw new ArgumentException("The array must have at least one element");
            NColumns = 1;
            _elements = new double[NRows][];
            for(int r = 0; r < NRows; r++)
            {
                _elements[r] = new double[1];
                _elements[r][0] = array[r];
            }
        }

        /// <summary>
        /// Creates a new matrix, nRows x nColumns, with all elements equal to zero.
        /// </summary>
        /// <param name="nRows">Number of rows.</param>
        /// <param name="nColumns">Number of columns.</param>
        public Matrix(int nRows, int nColumns)
        {
            NRows = nRows;
            NColumns = nColumns;
            _elements = new double[NRows][];
            for (int r = 0; r < NRows; r++)
                _elements[r] = new double[NColumns];
        }
        #endregion

        /// <summary>
        /// Creates an nxn identity matrix.
        /// </summary>
        /// <param name="n">Size of idenity matrix = number of rows = number of columns.</param>
        public static Matrix IdentityMatrix(int n)
        {
            Matrix M = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                M[i, i] = 1.0;
            return M;
        }


        /// <summary>
        /// Returns element [r,c] from the matrix.
        /// </summary>
        /// <param name="r">row index</param>
        /// <param name="c">column index</param>
        public double this[int r, int c]
        {
            get { return _elements[r][c]; }
            set { _elements[r][c] = value; }
        }

        public double[,] ToArray()
        {
            double[,] array = new double[NRows,NColumns];
            for (int r = 0; r < NRows; r++)
                for(int c = 0; c < NColumns; c++)
                    array[r, c] = _elements[r][c];
            return array;
        }

        public Matrix Copy()
        {
            Matrix M = new Matrix(NRows, NColumns);
            for (int r = 0; r < NRows; r++)
                for (int c = 0; c < NColumns; c++)
                    M[r, c] = _elements[r][c];
            return M;
        }

        /// <summary>
        /// Returns the sum of matrix M and the existing matrix.
        /// </summary>
        /// <param name="M">Must be the same size as the existing matrix.</param>
        public Matrix Add(Matrix M)
        {
            if (NRows != M.NRows || NColumns != M.NColumns)
                throw new ArgumentException("Cannot add. The matrices must be the same size.");
            Matrix sum = new Matrix(NRows, NColumns);
            for (int r = 0; r < NRows; r++)
                for (int c = 0; c < NColumns; c++)
                    sum[r, c] = this[r, c] + M[r, c];
            return sum;
        }

        /// <summary>
        /// Matrix subtraction. X - M.
        /// </summary>
        /// <param name="M">The matrix to be subtracted.</param>
        public Matrix Subtract(Matrix M)
        {
            Matrix minusM = M.Multiply(-1);
            return Add(minusM);
        }

        /// <summary>
        /// Returns the transpose of the existing matrix.
        /// </summary>
        public Matrix Transpose()
        {
            Matrix T = new Matrix(NColumns, NRows);
            for (int r = 0; r < NRows; r++)
                for (int c = 0; c < NColumns; c++)
                    T._elements[c][r] = _elements[r][c];
            return T;
        }

        /// <summary>
        /// Matrix multiplication. Post multiplies the existing matrix by another matrix, M.
        /// </summary>
        /// <param name="M">The number of rows in m should be equal to the number of columns in the existing matrix.</param>
        public Matrix Multiply(Matrix M)
        {
            if (NColumns != M.NRows)
                throw new ArgumentException("Cannot multiply. Number of columns in first matrix must equal the number of rows in the second.");

            Matrix product = new Matrix(NRows, M.NColumns);
            for (int r = 0; r < product.NRows; r++)
                for (int c = 0; c < product.NColumns; c++)
                {
                    double d = 0;
                    for (int i = 0; i < NColumns; i++)
                        d += _elements[r][i] * M._elements[i][c];   //NOTE: explicitly accessing the elements, M._elements[i, c], is faster than using get, M[i, c]. 
                    product._elements[r][c] = d;
                }
            return product;
        }

        /// <summary>
        /// Raises a matrix to the n-th power.
        /// </summary>
        public Matrix Power(int n)
        {
            Matrix M = Copy();
            for (int i = 0; i < n - 1; i++)
                M = M.Multiply(this);
            return M;
        }

        /// <summary>
        /// Scalar multiplication. Multiplies the existing matrix by d.
        /// </summary>
        /// <param name="d">The scalar by which the matrix is being multiplied.</param>
        /// <returns></returns>
        public Matrix Multiply(double d)
        {
            Matrix product = new Matrix(NRows, NColumns);
            for (int r = 0; r < product.NRows; r++)
                for (int c = 0; c < product.NColumns; c++)
                    product[r, c] = d * _elements[r][c];
            return product;
        }

        /// <summary>
        /// Kronecker product of two matrices, A and B.
        /// If A is m x n and B is p x q, then the return matrix will be mp x nq.
        /// </summary>
        public static Matrix KroneckerProduct(Matrix A, Matrix B)
        {
            Matrix M = new Matrix(A.NRows * B.NRows, A.NColumns * B.NColumns);
            for (int rowA = 0; rowA < A.NRows; rowA++)
            {
                for (int rowB = 0; rowB < B.NRows; rowB++)
                {
                    for (int colA = 0; colA < A.NColumns; colA++)
                    {
                        for (int colB = 0; colB < B.NColumns; colB++)
                        {
                            M[rowA * B.NRows + rowB, colA * B.NColumns + colB] = A[rowA, colA] * B[rowB, colB];
                        }
                    }
                }
            }
            return M;
        }

        /// <summary>
        /// For a matrix M, often noted vec(M).
        /// If M is a matrix of column vectors, M = [c1 c2 c3], then vec(M) would be a vector with c1 stacked on top of c2 on top of c3.
        /// </summary>
        public Matrix Vectorization()
        {
            Matrix M = new Matrix(NRows * NColumns, 1);
            for (int c = 0; c < NColumns; c++)
                for (int r = 0; r < NRows; r++)
                    M[c * NRows + r, 0] = this[r, c];
            return M;
        }

        /// <summary>
        /// Inverse function of matrix vectorization.
        /// </summary>
        /// <param name="nColumns">The number of columns of the resulting matrix.</param>
        /// <returns></returns>
        public Matrix VectorizationInverse(int nColumns)
        {
            if(NColumns != 1)
                throw new ArgumentException("To perform VectorizationInverse matrix must be a vector");
            int rem;
            int nRows = Math.DivRem(NRows, nColumns, out rem);
            if (rem != 0)
                throw new ArgumentException("Cannot evenly divide this vector into this many columns");
            Matrix M = new Matrix(nRows, nColumns);
            for (int c = 0; c < nColumns; c++)
                for (int r = 0; r < nRows; r++)
                    M[r, c] = this[c * nRows + r, 0];
            return M;
        }

        #region Matrix Decomposition
        /// <summary>
        /// Decomposes a symmetric square matrix into upper and lower triangular matrices, L and U, where L = U'.
        /// </summary>
        /// <param name="U">Upper triangular matrix</param>
        /// <param name="L">Lower triangular matrix</param>
        public void CholeskyDecomposition(out Matrix L, out Matrix U)
        {
            if (NRows != NColumns)
                throw new ArgumentException("Can only decompose square matrix. Number of rows and columns must be equal.");
            if (!IsSymmetric())
                throw new ArgumentException("Can only use Cholesky with a symmetric matrix.");
            int n = NRows;
            L = new Matrix(n, n);

            for(int r = 0; r < n; r++)
                for (int c = 0; c <= r; c++)
                {
                    if (r == c)
                    {
                        double d = this[r, c];
                        for (int i = 0; i < r; i++)
                            d -= L[r, i] * L[r, i];
                        L[r, c] = Math.Sqrt(d);
                    }
                    else
                    {
                        double d = this[r, c];
                        for (int i = 0; i < c; i++)
                            d -= L[r, i] * L[c, i];
                        d /= L[c, c];
                        L[r, c] = d;
                    }
                }

            U = L.Transpose();
        }

        /// <summary>
        /// Returns true if matrix is square and X[r, c] = X[c, r] for all r and c.
        /// </summary>
        /// <returns></returns>
        public bool IsSymmetric()
        {
            if (NRows != NColumns) return false;
            for (int r = 0; r < NRows; r++)
                for (int c = 0; c < r; c++)
                    if (Math.Abs(this[r, c] - this[c, r]) > double.Epsilon) return false;
            return true;
        }

        /// <summary>
        /// Decomposes a square matrix into upper and lower triangular matrices, L and U, such that X = LU.
        /// </summary>
        /// <param name="U">Upper triangular matrix</param>
        /// <param name="L">Lower triangular matrix</param>
        public void LuDecomposition(out Matrix L, out Matrix U)
        {
            if (NRows != NColumns)
                throw new ArgumentException("Can only decompose square matrix. Number of rows and columns must be equal.");

            int n = NRows;
            L = new Matrix(n, n);
            U = Copy();

            for (int r = 0; r < n; r++)
            {
                L[r, r] = U[r, r];
                for (int c = 0; c < n; c++)
                    U[r, c] = U[r, c] / L[r, r];
                
                for (int i = r + 1; i < n; i++)
                {
                    L[i, r] = U[i, r];
                    for (int j = 0; j < n; j++)
                        U[i, j] -= L[i, r] * U[r, j];
                }
            }
        }
        #endregion

        #region Matrix Inversion
        /// <summary>
        /// Used for regression analysis.
        /// Calculates the inverse of the existing matrix. Matrix must be square.
        /// If the matrix contains a diagonal element equal to zero, returns the inverse matrix of the remaining rows and columns, with zero rows and columns inserted.
        /// A diagonal element equal to zero happens when one of the columns of the X matrix is all zeros.
        /// </summary>
        public Matrix PseudoInverse()
        {
            if (NRows != NColumns)
                throw new ArgumentException("Can only invert square matrix. Number of rows and columns must be equal.");
            int nBig = NRows;

            List<int> zeroIndexes = new List<int>();
            for (int i = 0; i < nBig; i++)
                if(_elements[i][i] == 0)
                    zeroIndexes.Add(i);

            if (zeroIndexes.Count == 0)
                return Inverse();

            int nSmall = nBig - zeroIndexes.Count;
            Matrix M = new Matrix(nSmall, nSmall);
            int newRowIndex, newColIndex = 0;
            for (int c = 0; c < NColumns; c++)
            {
                if(zeroIndexes.Contains(c)) continue;
                newRowIndex = 0;
                for (int r = 0; r < NRows; r++)
                {
                    if (zeroIndexes.Contains(r)) continue;
                    M._elements[newRowIndex][newColIndex] = _elements[r][c];
                    newRowIndex++;
                }
                newColIndex++;
            }

            Matrix MInv = M.Inverse();
            Matrix Inv = new Matrix(NRows, NColumns);
            newRowIndex = 0;
            for (int r = 0; r < NRows; r++)
            {
                if (zeroIndexes.Contains(r)) continue;  //leave row all zeros.
                newColIndex = 0;
                for (int c = 0; c < NColumns; c++)
                {
                    if (zeroIndexes.Contains(c))
                        Inv._elements[r][c] = 0;
                    else
                    {
                        Inv._elements[r][c] = MInv._elements[newRowIndex][newColIndex];
                        newColIndex++;
                    }
                }
                newRowIndex++;
            }
            return Inv;
        }

        /// <summary>
        /// Calculates the inverse of the existing matrix. Matrix must be square.
        /// </summary>
        public Matrix Inverse()
        {
            //This method first decomposes the existing matrix into a lower and upper tringular matrix. These
            //can be more easily inverted.
            //There may be more precise or efficient algorithms, but this is very easy to follow.

            if (NRows != NColumns)
                throw new ArgumentException("Can only invert square matrix. Number of rows and columns must be equal.");

            Matrix L, U;
            LuDecomposition(out L, out U);

            Matrix Linv = L.LowerTriangularInverse();
            Matrix Uinv = U.Transpose().LowerTriangularInverse().Transpose();

            Matrix inverse = Uinv.Multiply(Linv);

            return inverse;
        }

        public bool IsCloseToIdentity(int precision)
        {
            if (NRows != NColumns) return false;
            double d = Math.Pow(10, -precision);
            for(int r = 0; r<NRows; r++)
                for (int c = 0; c < NColumns; c++)
                {
                    if (r == c)
                    {
                        if (Math.Abs(this[r, c] - 1) > d) return false;
                    }
                    else
                    {
                        if (Math.Abs(this[r, c]) > d) return false;
                    }
                }
            return true;
        }

        public bool IsLowerTriangular()
        {
            return Transpose().IsUpperTriangular();
        }

        public bool IsUpperTriangular()
        {
            if (NRows != NColumns)
                throw new ArgumentException("Upper and lower triangular only defined for square matrices. Number of rows and columns must be equal.");
            int n = NRows;
            for (int r = 1; r < n; r++)
            {
                for (int c = 0; c < r; c++)
                    if (this[r, c] != 0)
                        return false;
            }
            return true;
        }

        /// <summary>
        /// Calculates the inverse of a lower triangular matrix. Used to calculate general inverses.
        /// </summary>
        /// <returns></returns>
        private Matrix LowerTriangularInverse()
        {
            if (!IsLowerTriangular())
                throw new Exception("Can only use with lower triangular matrix");
            int n = NRows;

            Matrix D = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                D[i, i] = 1.0 / _elements[i][i];
            Matrix C = D.Multiply(this);

            //needed to prevent occasional rounding error
            for (int i = 0; i < n; i++)
                C[i, i] = 1.0; 
            
            Matrix Cinv = C.LowerTriangularInverseOneDiagonal();
            Matrix M = Cinv.Multiply(D);

            return M;
        }

        /// <summary>
        /// Used to calculate LowerTriangularInverse, which is used to calculate inverses.
        /// </summary>
        private Matrix LowerTriangularInverseOneDiagonal()
        {
            //The inverse of a lower triangular matrix with ones on the diagonal is also
            //lower triangular with ones on the diagonal.
            //This problem is relatively easy to solve relative and can be used to solve the more
            //general inversion problem.

            if (!IsLowerTriangular())
                throw new Exception("Can only use with lower triangular matrix");
            int n = NRows;
            for (int i = 0; i < n; i++)
                if (_elements[i][i] != 1)
                    throw new Exception("Can only use where all diagonal elements are 1");

            Matrix M = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                M[i, i] = 1;

            for(int d = 1; d < n; d++)
                for (int r = d; r < n; r++)
                {
                    int c = r - d;
                    for (int i = 0; i < r; i++)
                        M._elements[r][c] -= _elements[r][i] * M._elements[i][c];
                }

            return M;
        }
        #endregion

    }

}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.